// $Id: Reference.cpp,v 1.35 2000/04/10 16:01:35 chuang Exp $
#include <string.h>
#include "ComObject.h"
#include "Reference.h"

Reference::Reference (IDispatch *pDispatch, const Interface *pInterface):
    m_pDispatch(pDispatch),
    m_pInterface(pInterface),
    m_pConnectionPoint(0)
{ }

Reference::~Reference()
{
    if (m_pConnectionPoint != 0) {
        m_pConnectionPoint->Unadvise(m_adviseCookie);
        m_pConnectionPoint->Release();
    }
    m_pDispatch->Release();
}

void
Reference::advise (Tcl_Interp *interp,
                   const Interface &sourceInterfaceDesc,
                   TclObject servant)
{
    HRESULT hr;

    // Get connection point container.
    IConnectionPointContainer *pContainer;
    hr = m_pDispatch->QueryInterface(
        IID_IConnectionPointContainer,
        reinterpret_cast<void **>(&pContainer));
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }

    // Find connection point.
    hr = pContainer->FindConnectionPoint(
        sourceInterfaceDesc.iid(), &m_pConnectionPoint);
    pContainer->Release();
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }

    // Create event sink.
    IUnknown *pSink = ComObject::newInstance(
        sourceInterfaceDesc,
        Tcl_GetCurrentThread(),
        interp,
        servant,
        "",
        true);

    // Will unadvise when the reference is destroyed.
    hr = m_pConnectionPoint->Advise(pSink, &m_adviseCookie);
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }
}

const Interface *
Reference::interfaceDesc ()
{
    if (m_pInterface != 0) {
        return m_pInterface;
    }

    HRESULT hr;
    
    // Ask the IDispatch interface for type information.
    unsigned count;
    hr = m_pDispatch->GetTypeInfoCount(&count);
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }
    if (count == 0) {
        return 0;
    }

    ITypeInfo *pITypeInfo;
    hr = m_pDispatch->GetTypeInfo(
        0, 
        LOCALE_USER_DEFAULT,
        &pITypeInfo);
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }

    // Get the interface description.
    TypeInfo typeInfo(pITypeInfo);
    TypeAttr typeAttr(typeInfo);
#if 0
    RPC_STATUS rpcStatus;
    if (UuidIsNil(&typeAttr->guid, &rpcStatus)) {
        m_pInterface = new Interface(typeInfo);
        return m_pInterface;
    }
#endif
    m_pInterface = InterfaceManager::instance().newInterface(
        typeAttr->guid, typeInfo);

    // Get a pointer to the derived interface.
    IDispatch *pNew;
    hr = m_pDispatch->QueryInterface(
        typeAttr->guid,
        reinterpret_cast<void **>(&pNew));
    if (SUCCEEDED(hr)) {
        m_pDispatch->Release();
        m_pDispatch = pNew;
    }

    return m_pInterface;
}

HRESULT
Reference::invoke (MEMBERID memberid,
                   INVOKEKIND invokeKind,
                   const Arguments &arguments,
                   _variant_t *pResult)

{
    HRESULT hr;

    WORD flags;
    switch (invokeKind) {
    case INVOKE_PROPERTYGET:
        flags = DISPATCH_PROPERTYGET;
        break;

    case INVOKE_PROPERTYPUT:
        flags = DISPATCH_PROPERTYPUT;
        break;

    case INVOKE_PROPERTYPUTREF:
        flags = DISPATCH_PROPERTYPUTREF;
        break;

    default:
        flags = DISPATCH_METHOD;
        break;
    }

    EXCEPINFO excepInfo;
    unsigned argErr;

    if (interfaceDesc()->dispatchable()) {
        // Invoke through IDispatch interface.
        hr = m_pDispatch->Invoke(
            memberid,
            IID_NULL,
            LOCALE_USER_DEFAULT,
            flags,
            arguments.dispParams(),
            pResult,
            &excepInfo,
            &argErr);
    } else {
        // Invoke through virtual function table.
        ITypeInfo *pTypeInfo = interfaceDesc()->typeInfo();
        hr = pTypeInfo->Invoke(
            m_pDispatch,
            memberid,
            flags,
            arguments.dispParams(),
            pResult,
            &excepInfo,
            &argErr);
    }

    if (hr == DISP_E_EXCEPTION) {
        // Clean up exception information strings.
        SysFreeString(excepInfo.bstrSource);
        SysFreeString(excepInfo.bstrDescription);
        SysFreeString(excepInfo.bstrHelpFile);

        hr = excepInfo.scode;
    }

    return hr;
}

Singleton<ReferenceFactory> ReferenceFactory::ms_singleton;

ReferenceFactory &
ReferenceFactory::instance ()
{
    return ms_singleton.instance();
}

void
ReferenceFactory::insert (Reference *pRef)
{
    LOCK_MUTEX(m_mutex)

    m_history.insert(pRef);
}

Reference *
ReferenceFactory::createObject (REFCLSID clsid,
                                const Interface *pInterface,
                                const char *serverHost)
{
    // If we know it's a custom interface, then query for an interface pointer
    // to that interface, otherwise query for an IDispatch interface.
    const IID &iid = (pInterface == 0) ? IID_IDispatch : pInterface->iid();

    HRESULT hr;
    IDispatch *pDispatch;

#ifdef _WIN32_DCOM
    // Create the instance of the object
    _bstr_t serverHostBstr(serverHost);

    COSERVERINFO serverInfo, *pServerInfo;
    if (serverHost == 0
     || serverHost[0] == '\0'
     || strcmp(serverHost, "localhost") == 0
     || strcmp(serverHost, "127.0.0.1") == 0) {
        pServerInfo = NULL;
    } else {
        pServerInfo = &serverInfo;
        memset(&serverInfo, 0, sizeof(serverInfo));
        serverInfo.pwszName = serverHostBstr;
    }

    MULTI_QI qi;
    qi.pIID = &iid;
    qi.pItf = NULL;
    qi.hr = 0;

    hr = CoCreateInstanceEx(
        clsid,
        NULL,
        CLSCTX_SERVER,
        pServerInfo,
        1,
        &qi);

    if (SUCCEEDED(hr)) {
        pDispatch = static_cast<IDispatch *>(qi.pItf);
    }

#else
    hr = CoCreateInstance(
        clsid,
        NULL,
        CLSCTX_SERVER,
        iid,
        reinterpret_cast<void **>(&pDispatch));
#endif

    if (FAILED(hr)) {
        _com_issue_error(hr);
    }

    Reference *pRef = new Reference(pDispatch, pInterface);
    insert(pRef);
    return pRef;
}

Reference *
ReferenceFactory::createObject (const char *progId, const char *serverHost)
{
    // Convert the Prog ID to a CLSID.
    CLSID clsid;
    HRESULT hr = CLSIDFromProgID(_bstr_t(progId), &clsid);
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }

    return createObject(clsid, 0, serverHost);
}

Reference *
ReferenceFactory::getObject (REFCLSID clsid,
                             const Interface *pInterface)
{
    HRESULT hr;

    // Retrieve the instance of the object.
    IUnknown *pUnknown;
    hr = GetActiveObject(clsid, NULL, &pUnknown);
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }

    // Query the interface pointer.
    IDispatch *pDispatch;
    hr = pUnknown->QueryInterface(
        (pInterface == 0) ? IID_IDispatch : pInterface->iid(),
        reinterpret_cast<void **>(&pDispatch));
    pUnknown->Release();
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }

    Reference *pRef = new Reference(pDispatch, pInterface);
    insert(pRef);
    return pRef;
}

Reference *
ReferenceFactory::getObject (const char *progId)
{
    // Convert the Prog ID to a CLSID.
    CLSID clsid;
    HRESULT hr = CLSIDFromProgID(_bstr_t(progId), &clsid);
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }

    return getObject(clsid, 0);
}

Reference *
ReferenceFactory::newReference (IUnknown *pUnknown,
                                REFIID iid)
{
    IDispatch *pDispatch;
    HRESULT hr = pUnknown->QueryInterface(
        iid,
        reinterpret_cast<void **>(&pDispatch));
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }

    Reference *pRef = new Reference(pDispatch, 0);
    insert(pRef);
    return pRef;
}

Reference *
ReferenceFactory::newReference (IUnknown *pUnknown,
                                const Interface *pInterface)
{
    // If we know it's a custom interface, then query for an interface pointer
    // to that interface, otherwise query for an IDispatch interface.
    const IID &iid = (pInterface == 0) ? IID_IDispatch : pInterface->iid();

    IDispatch *pDispatch;
    HRESULT hr = pUnknown->QueryInterface(
        iid,
        reinterpret_cast<void **>(&pDispatch));
    if (FAILED(hr)) {
        _com_issue_error(hr);
    }

    Reference *pRef = new Reference(pDispatch, pInterface);
    insert(pRef);
    return pRef;
}

void
ReferenceFactory::destroy (Reference *pRef)
{
    LOCK_MUTEX(m_mutex)

    m_history.erase(pRef);
    delete pRef;
}

void
ReferenceFactory::clear ()
{
    for (History::iterator p = m_history.begin(); p != m_history.end(); ++p) {
        delete *p;
    }
    m_history.clear();
}
