// $Id: Reference.h,v 1.20 2000/04/22 21:39:36 chuang Exp $
#ifndef REFERENCE_H
#define REFERENCE_H

#pragma warning(disable: 4786)
#include <set>
#include "tcomApi.h"
#include "TclObject.h"
#include "TypeInfo.h"
#include "Arguments.h"
#include "Singleton.h"

// This class holds an interface pointer and the interface description needed
// to invoke methods on it.

class TCOM_API Reference
{
    friend class ReferenceFactory;

    // interface pointer to the object
    IDispatch *m_pDispatch;

    // interface description includes information about methods and properties
    const Interface *m_pInterface;

    // interface pointer to connection point
    IConnectionPoint *m_pConnectionPoint;

    // cookie returned from Advise
    DWORD m_adviseCookie;

    // We assume the reference count on the interface pointer has already
    // been incremented.  This object will decrement the reference count when
    // it is destroyed.
    Reference(IDispatch *pDispatch, const Interface *pInterface);

    // Release the interface pointer.
    ~Reference();

    // Do not allow instances of this class to be copied.
    Reference(const Reference &rhs);
    Reference &operator=(const Reference &rhs);

public:
    // Get interface description.
    const Interface *interfaceDesc();

    // Invoke a method
    HRESULT invoke(
        MEMBERID memberid,
        INVOKEKIND invokeKind,
        const Arguments &arguments,
        _variant_t *pResult);

    // Get raw interface pointer.
    IDispatch *dispatch () const
    { return m_pDispatch; }

    // Create an event sink and connect to the event source.
    void advise(
        Tcl_Interp *interp,
        const Interface &sourceInterfaceDesc,
        TclObject servant);
};

// This class manages the life cycle of object references.

class TCOM_API ReferenceFactory
{
    DECLARE_MUTEX(m_mutex)

    // Remember references that have not been released yet.  This allows us to
    // clean up references that the scripter neglected to release.
    typedef std::set<Reference *> History;
    History m_history;

    friend class Singleton<ReferenceFactory>;
    static Singleton<ReferenceFactory> ms_singleton;

    // Do not allow others to create or copy instances of this class.
    ReferenceFactory ()
    { }

    ReferenceFactory(const ReferenceFactory &); // not implemented
    void operator=(const ReferenceFactory &);   // not implemented

    // Add reference to history.
    void insert(Reference *pRef);

public:
    ~ReferenceFactory ()
    { clear(); }

    // Get singleton instance.
    static ReferenceFactory &instance();

    // Perform a QueryInterface on the interface pointer and create a reference.
    Reference *newReference(IUnknown *pUnknown, REFIID iid);

    // Perform a QueryInterface on the interface pointer and create a reference.
    Reference *newReference(IUnknown *pUnknown, const Interface *pInterface=0);

    // Create an object using CoCreateInstance and construct a reference.
    Reference *createObject(
        REFCLSID clsid, const Interface *pInterface, const char *serverHost);

    // Create an object using CoCreateInstance and construct a reference.
    Reference *createObject(const char *progId, const char *serverHost);

    // Get an object using GetActiveObject and construct a reference.
    Reference *getObject(REFCLSID clsid, const Interface *pInterface);

    // Get an object using GetActiveObject and construct a reference.
    Reference *getObject(const char *progId);

    // Destroy reference.
    void destroy(Reference *pRef);

    // Release any existing references.
    void clear();
};

#endif 
