// $Id: TclObject.cpp,v 1.4 2000/04/06 17:07:56 chuang Exp $
#include "TclObject.h"
#ifdef WIN32
#include "refCmd.h"
#include "naCmd.h"
#include "nullCmd.h"
#include "interfacePointer.h"
#endif

TclObject::TclObject ():
    m_pObj(Tcl_NewObj())
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (const TclObject &rhs):
    m_pObj(rhs.m_pObj)
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (Tcl_Obj *pObj):
    m_pObj(pObj)
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (const char *src, int len):
    m_pObj(Tcl_NewStringObj(src, len))
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (const std::string &s):
    m_pObj(Tcl_NewStringObj(s.data(), s.size()))
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (bool value):
    m_pObj(Tcl_NewBooleanObj(static_cast<int>(value)))
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (int value):
    m_pObj(Tcl_NewIntObj(value))
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (long value):
    m_pObj(Tcl_NewLongObj(value))
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (double value):
    m_pObj(Tcl_NewDoubleObj(value))
{ Tcl_IncrRefCount(m_pObj); }

TclObject::TclObject (int objc, Tcl_Obj *CONST objv[]):
    m_pObj(Tcl_NewListObj(objc, objv))
{ Tcl_IncrRefCount(m_pObj); }

TclObject::~TclObject ()
{ Tcl_DecrRefCount(m_pObj); }

TclObject &
TclObject::operator= (const TclObject &rhs)
{
    Tcl_IncrRefCount(rhs.m_pObj);
    Tcl_DecrRefCount(m_pObj);
    m_pObj = rhs.m_pObj;
    return *this;
}

TclObject &
TclObject::operator= (Tcl_Obj *pObj)
{
    Tcl_IncrRefCount(pObj);
    Tcl_DecrRefCount(m_pObj);
    m_pObj = pObj;
    return *this;
}

bool
TclObject::getBool () const
{
    int value;
    Tcl_GetBooleanFromObj(0, m_pObj, &value);
    return value != 0;
}

int
TclObject::getInt () const
{
    int value;
    Tcl_GetIntFromObj(0, m_pObj, &value);
    return value;
}

long
TclObject::getLong () const
{
    long value;
    Tcl_GetLongFromObj(0, m_pObj, &value);
    return value;
}

double
TclObject::getDouble () const
{
    double value;
    Tcl_GetDoubleFromObj(0, m_pObj, &value);
    return value;
}

TclObject &
TclObject::lappend (Tcl_Obj *pElement)
{
    if (Tcl_IsShared(m_pObj)) {
        Tcl_DecrRefCount(m_pObj);
        m_pObj = Tcl_DuplicateObj(m_pObj);
        Tcl_IncrRefCount(m_pObj);
    }
    Tcl_ListObjAppendElement(NULL, m_pObj, pElement);
    // TODO: Should check for error result if conversion to list failed.
    return *this;
}

#ifdef WIN32

TclObject::TclObject (_variant_t &var, const Type &type, Tcl_Interp *interp)
{
    HRESULT hResult;

    if (var.vt & VT_ARRAY) {
        // We can handle only one-dimensional arrays.
        SAFEARRAY *psa = var.parray;
        if (SafeArrayGetDim(psa) != 1) {
            _com_issue_error(E_INVALIDARG);
        }

        // Get number of elements.
        long uBound;
        hResult = SafeArrayGetUBound(psa, 1, &uBound);
        if (FAILED(hResult)) {
            _com_issue_error(hResult);
        }
        long length = uBound + 1;

        // Get element type.
        VARTYPE vt = var.vt & VT_TYPEMASK;

        switch (vt) {
        case VT_UNKNOWN:
            // Convert array of IUnknown to Tcl list of interface pointer
            // handles.
            {
                IUnknown **pData;
                hResult = SafeArrayAccessData(
                    psa,
                    reinterpret_cast<void **>(&pData));
                if (FAILED(hResult)) {
                    _com_issue_error(hResult);
                }

                m_pObj = Tcl_NewListObj(0, 0);
                for (int i = 0; i < length; ++i) {
                    Tcl_Obj *pElement =
                        HandleSupport<Reference>::instance(interp).newObj(
                            ReferenceFactory::instance().newReference(
                                pData[i]));
                    Tcl_ListObjAppendElement(interp, m_pObj, pElement);
                }

                hResult = SafeArrayUnaccessData(psa);
                if (FAILED(hResult)) {
                    _com_issue_error(hResult);
                }
            }
            break;

        case VT_UI1:
            {
                unsigned char *pData;
                hResult = SafeArrayAccessData(
                    psa,
                    reinterpret_cast<void **>(&pData));
                if (FAILED(hResult)) {
                    _com_issue_error(hResult);
                }

                m_pObj =
#if TCL_MINOR_VERSION >= 1
                    // Convert array of bytes to Tcl byte array.
                    Tcl_NewByteArrayObj(pData, length);
#else
                    // Convert array of bytes to Tcl string.
                    Tcl_NewStringObj(reinterpret_cast<char *>(pData), length);
#endif

                hResult = SafeArrayUnaccessData(psa);
                if (FAILED(hResult)) {
                    _com_issue_error(hResult);
                }
            }
            break;

        default:
            // Convert array of other types to Tcl list of objects.
            {
                m_pObj = Tcl_NewListObj(0, 0);
                for (long i = 0; i < length; ++i) {
                    _variant_t elementVar;

                    if (vt == VT_VARIANT) {
                        hResult = SafeArrayGetElement(psa, &i, &elementVar);
                    } else {
                        // I hope the element can be contained in a VARIANT.
                        elementVar.ChangeType(vt);
                        hResult = SafeArrayGetElement(
                            psa,
                            &i,
                            &elementVar.punkVal);
                    }
                    if (FAILED(hResult)) {
                        _com_issue_error(hResult);
                    }

                    TclObject element(elementVar, type, interp);
                    Tcl_ListObjAppendElement(interp, m_pObj, element);
                }
            }
        }

    } else {
        switch (var.vt) {
        case VT_BOOL:
            m_pObj = Tcl_NewBooleanObj(static_cast<bool>(var));
            break;

        case VT_INT:
        case VT_I1:
        case VT_I2:
        case VT_I4:
        case VT_UINT:
        case VT_UI1:
        case VT_UI2:
        case VT_UI4:
            m_pObj = Tcl_NewLongObj(var);
            break;

        case VT_R4:
        case VT_R8:
            m_pObj = Tcl_NewDoubleObj(var);
            break;

        case VT_DATE:
            {
                double days;
                VarR8FromDate(var, &days);
                m_pObj = Tcl_NewLongObj((days - 25570) * 86400);
            }
            break;

        case VT_DISPATCH:
            {
                const Interface *pInterface =
                    InterfaceManager::instance().find(type.iid());
                m_pObj = HandleSupport<Reference>::instance(interp).newObj(
                    ReferenceFactory::instance().newReference(
                        V_DISPATCH(&var),
                        pInterface));
            }
            break;

        case VT_UNKNOWN:
            {
                const Interface *pInterface =
                    InterfaceManager::instance().find(type.iid());
                m_pObj = HandleSupport<Reference>::instance(interp).newObj(
                    ReferenceFactory::instance().newReference(
                        V_UNKNOWN(&var),
                        pInterface));
            }
            break;

        case VT_NULL:
            m_pObj = Tcl_NewObj();
            break;

        default:
            {
                _bstr_t str(var);
#if TCL_MINOR_VERSION >= 2
                // Uses Unicode function introduced in Tcl 8.2.
                m_pObj = Tcl_NewUnicodeObj(str, str.length());
#else
                m_pObj = Tcl_NewStringObj(str, str.length());
#endif
            }
        }
    }

    Tcl_IncrRefCount(m_pObj);
}

_variant_t
TclObject::toVariant (VARTYPE vt, Tcl_Interp *interp, bool isOutValue)
{
    if (interp != 0
     && HandleSupport<Reference>::instance(interp).count(m_pObj) > 0) {
        // Convert interface pointer handle to interface pointer.
        Reference *pReference =
            HandleSupport<Reference>::instance(interp).find(m_pObj);
        if (isOutValue) {
            // Must increment reference count of interface pointers returned
            // from methods.
            pReference->dispatch()->AddRef();
        }
        return _variant_t(pReference->dispatch());

    } else if (m_pObj->typePtr == &interfacePointerObjType) {
        // Convert to interface pointer.
        IUnknown *pUnknown = reinterpret_cast<IUnknown *>(
            m_pObj->internalRep.otherValuePtr);
        if (isOutValue) {
            // Must increment reference count of interface pointers returned
            // from methods.
            pUnknown->AddRef();
        }
        return _variant_t(pUnknown);

    } else if (m_pObj->typePtr == Tcl_GetObjType("list")) {
        // Convert Tcl list to array of VARIANT.
        int length;
        Tcl_ListObjLength(interp, m_pObj, &length);

        SAFEARRAYBOUND saBound[1];
        saBound[0].lLbound = 0;
        saBound[0].cElements = length;

        SAFEARRAY *psa = SafeArrayCreate(VT_VARIANT, 1, saBound);
        if (psa == 0) {
            _com_issue_error(E_OUTOFMEMORY);
        }

        _variant_t *pData;
        HRESULT hResult;
        hResult = SafeArrayAccessData(psa, reinterpret_cast<void **>(&pData));
        if (FAILED(hResult)) {
            _com_issue_error(hResult);
        }

        for (int i = 0; i < length; ++i) {
            Tcl_Obj *pElement;
            Tcl_ListObjIndex(interp, m_pObj, i, &pElement);
            TclObject value(pElement);
            pData[i] = value.toVariant(vt, interp, isOutValue);
        }

        hResult = SafeArrayUnaccessData(psa);
        if (FAILED(hResult)) {
            _com_issue_error(hResult);
        }

        _variant_t var;
        var.vt = VT_ARRAY | VT_VARIANT;
        var.parray = psa;
        return var;

#if TCL_MINOR_VERSION >= 1
    } else if (m_pObj->typePtr == Tcl_GetObjType("bytearray")) {
        // Convert Tcl byte array to array of bytes.
        int length;
        unsigned char *m_pObjData = Tcl_GetByteArrayFromObj(m_pObj, &length);

        SAFEARRAYBOUND saBound[1];
        saBound[0].lLbound = 0;
        saBound[0].cElements = length;

        SAFEARRAY *psa = SafeArrayCreate(VT_UI1, 1, saBound);
        if (psa == 0) {
            _com_issue_error(E_OUTOFMEMORY);
        }

        unsigned char *pDestData;
        HRESULT hResult;
        hResult = SafeArrayAccessData(
            psa,
            reinterpret_cast<void **>(&pDestData));
        if (FAILED(hResult)) {
            _com_issue_error(hResult);
        }

        memcpy(pDestData, m_pObjData, length);

        hResult = SafeArrayUnaccessData(psa);
        if (FAILED(hResult)) {
            _com_issue_error(hResult);
        }

        _variant_t var;
        var.vt = VT_ARRAY | VT_UI1;
        var.parray = psa;
        return var;
#endif

    } else if (m_pObj->typePtr == &naObjType) {
        // This variant indicates a missing optional argument.
        return vtMissing;

    } else if (m_pObj->typePtr == &nullObjType && vt == VT_BSTR) {
        BSTR str = 0;
        return _variant_t(str);

    } else if (m_pObj->typePtr == Tcl_GetObjType("int")) {
        long value;
        if (Tcl_GetLongFromObj(interp, m_pObj, &value) != TCL_OK) {
            value = 0;
        }
        _variant_t var(value);
        if (vt != VT_VARIANT && vt != VT_USERDEFINED) {
            var.ChangeType(vt);
        }
        return var;

    } else if (m_pObj->typePtr == Tcl_GetObjType("double")) {
        double value;
        if (Tcl_GetDoubleFromObj(interp, m_pObj, &value) != TCL_OK) {
            value = 0.0;
        }
        _variant_t var(value);
        if (vt != VT_VARIANT && vt != VT_USERDEFINED) {
            var.ChangeType(vt);
        }
        return var;

    } else if (m_pObj->typePtr == Tcl_GetObjType("boolean")) {
        int value;
        if (Tcl_GetBooleanFromObj(interp, m_pObj, &value) != TCL_OK) {
            value = 0;
        }
        _variant_t var((value != 0) ? true : false);
        if (vt != VT_VARIANT && vt != VT_USERDEFINED) {
            var.ChangeType(vt);
        }
        return var;

    } else {
#if TCL_MINOR_VERSION >= 2
        // Uses Unicode function introduced in Tcl 8.2.
        Tcl_UniChar *pStringRep = Tcl_GetUnicode(m_pObj);
#else
        char *pStringRep = Tcl_GetStringFromObj(m_pObj, 0);
#endif
        _variant_t var(pStringRep);

        // Try to convert from a string representation.
        if (vt != VT_VARIANT && vt != VT_USERDEFINED) {
            var.ChangeType(vt);
        }
        return var;
    }
}

#endif
