// $Id: TypeInfo.h,v 1.25 2000/05/08 17:34:29 chuang Exp $
#ifndef TYPEINFO_H
#define TYPEINFO_H

#pragma warning(disable: 4786)
#include <vector>
#include <string>
#include <comdef.h>
#include "tcomApi.h"
#include "HashTable.h"
#include "Singleton.h"

// This wrapper class takes ownership of the resource passed into the
// constructor and is responsible for releasing it.

class TCOM_API TypeInfo
{
    ITypeInfo *m_pTypeInfo;

public:
    explicit TypeInfo (ITypeInfo *pTypeInfo=0):
        m_pTypeInfo(pTypeInfo)
    { }

    TypeInfo (const TypeInfo &rhs):
        m_pTypeInfo(rhs.m_pTypeInfo)
    { m_pTypeInfo->AddRef(); }

    ~TypeInfo ()
    {
        if (m_pTypeInfo != 0) {
            m_pTypeInfo->Release();
        }
    }

    TypeInfo &operator= (const TypeInfo &rhs)
    {
        if (this != &rhs) {
            m_pTypeInfo->Release();
            m_pTypeInfo = rhs.m_pTypeInfo;
            m_pTypeInfo->AddRef();
        }
        return *this;
    }

    // Get the raw pointer.
    ITypeInfo *get () const
    { return m_pTypeInfo; }

    // Dereference pointer.
    ITypeInfo *operator-> () const
    { return m_pTypeInfo; }

    // Get type info name.
    std::string name(MEMBERID id=MEMBERID_NIL);

    // Assuming this type is a class, get interface implemented by this class.
    ITypeInfo *getImplType(int bits);
};

// This wrapper class takes ownership of the resource and is responsible for
// releasing it.

class TCOM_API TypeAttr
{
    const TypeInfo &m_typeInfo;
    TYPEATTR *m_pTypeAttr;

    // Do not allow others to copy instances of this class.
    TypeAttr(const TypeAttr &);         // not implemented
    void operator=(const TypeAttr &);   // not implemented

public:
    // TODO: I should probably check for a failure result and throw an
    // exception in that case.
    TypeAttr (const TypeInfo &typeInfo):
        m_typeInfo(typeInfo)
    { m_typeInfo->GetTypeAttr(&m_pTypeAttr); }

    ~TypeAttr ()
    { m_typeInfo->ReleaseTypeAttr(m_pTypeAttr); }

    // Deference pointer.
    TYPEATTR *operator-> () const
    { return m_pTypeAttr; }
};

// This wrapper class takes ownership of the resource and is responsible for
// releasing it.

class TCOM_API FuncDesc
{
    TypeInfo &m_typeInfo;
    FUNCDESC *m_pFuncDesc;

    // Do not allow others to copy instances of this class.
    FuncDesc(const FuncDesc &);         // not implemented
    void operator=(const FuncDesc &);   // not implemented

public:
    FuncDesc(TypeInfo &typeInfo, unsigned index);

    ~FuncDesc ()
    { m_typeInfo->ReleaseFuncDesc(m_pFuncDesc); }

    // Dereference pointer.
    FUNCDESC *operator-> () const
    { return m_pFuncDesc; }
};

// This wrapper class takes ownership of the resource and is responsible for
// releasing it.

class TCOM_API VarDesc
{
    TypeInfo &m_typeInfo;
    VARDESC *m_pVarDesc;

    // Do not allow others to copy instances of this class.
    VarDesc(const VarDesc &);           // not implemented
    void operator=(const VarDesc &);    // not implemented

public:
    VarDesc(TypeInfo &typeInfo, unsigned index);

    ~VarDesc ()
    { m_typeInfo->ReleaseVarDesc(m_pVarDesc); }

    // Dereference pointer.
    VARDESC *operator-> () const
    { return m_pVarDesc; }
};

// This class describes a type for a function return value or parameter.

class TCOM_API Type
{
    std::string m_name;
    VARTYPE m_vt;
    IID m_iid;
    unsigned m_pointerCount;

    void readTypeDesc(TypeInfo &typeInfo, TYPEDESC &typeDesc);

public:
    Type (TypeInfo &typeInfo, TYPEDESC &typeDesc):
        m_pointerCount(0)
    { readTypeDesc(typeInfo, typeDesc); }

    Type(const std::string &str);

    const std::string &name () const
    { return m_name; }

    VARTYPE vartype () const
    { return m_vt; }

    const IID &iid () const
    { return m_iid; }

    unsigned pointerCount () const
    { return m_pointerCount; }

    // Get string representation.
    std::string toString() const;
};

// This class describes a function parameter.

class TCOM_API Parameter
{
    std::string m_name;
    Type m_type;
    unsigned m_flags;

public:
    Parameter(TypeInfo &typeInfo, ELEMDESC &elemDesc, const char *name);

    Parameter(
        const std::string &modes,
        const std::string &type,
        const std::string &name);

    Parameter (unsigned flags, const Type &type, const char *name):
        m_flags(flags),
        m_type(type),
        m_name(name)
    { }

    const std::string &name () const
    { return m_name; }

    const Type &type () const
    { return m_type; }

    unsigned flags () const
    { return m_flags; }
};

// This class describes a method.

class TCOM_API Method
{
public:
    typedef std::vector<Parameter> Parameters;

private:
    std::string m_name;
    Type m_type;
    Parameters m_parameters;
    MEMBERID m_memberid;
    INVOKEKIND m_invokeKind;
    short m_vtblIndex;      // position in virtual function table

protected:
    Method(TypeInfo &typeInfo, DISPID memberid, ELEMDESC &elemDesc);

public:
    Method(TypeInfo &typeInfo, FuncDesc &funcDesc);

    Method(
        MEMBERID memberid,
        const std::string &type,
        const std::string &name,
        INVOKEKIND invokeKind=INVOKE_FUNC);

    // Get method name.
    const std::string &name () const
    { return m_name; }

    // Get return value type.
    const Type &type () const
    { return m_type; }

    // Set return value type.
    void type (const Type &rhs)
    { m_type = rhs; }

    // Insert parameter.
    void addParameter (const Parameter &parameter)
    { m_parameters.push_back(parameter); }

    // Get parameters.
    const Parameters &parameters () const
    { return m_parameters; }

    // Get member ID.
    MEMBERID memberid () const
    { return m_memberid; }

    // Get indicator whether this is a method or property.
    INVOKEKIND invokeKind () const
    { return m_invokeKind; }

    // Set indicator whether this is a method or property.
    void invokeKind (INVOKEKIND invokeKind)
    { m_invokeKind = invokeKind; }

    // Get position in virtual function table.
    short vtblIndex () const
    { return m_vtblIndex; }
};

// This class describes a property.

class TCOM_API Property: public Method
{
    bool m_readOnly;

public:
    Property(TypeInfo &typeInfo, FuncDesc &funcDesc);
    Property(TypeInfo &typeInfo, VarDesc &varDesc);

    Property(
        MEMBERID memberid,
        const std::string &modes,
        const std::string &type,
        const std::string &name);

    // Get read only flag.
    bool readOnly () const
    { return m_readOnly; }

    // Set read only flag.
    void readOnly (bool value)
    { m_readOnly = value; }
};

// This describes an interface.

class TCOM_API Interface
{
    // Make the following a friend so it can construct instances of this class.
    friend class InterfaceManager;

public:
    typedef std::vector<Method> Methods;
    typedef std::vector<Property> Properties;

private:
    TypeInfo m_typeInfo;
    IID m_iid;
    std::string m_name;
    Methods m_methods;
    Properties m_properties;
    bool m_dispatchable;

    // Get information on methods and properties.
    void readFunctions(TypeInfo &typeInfo, TypeAttr &typeAttr);

    // Get information on interface.
    void readTypeInfo(TypeInfo &typeInfo);

    // Constructors
    Interface (TypeInfo &typeInfo):
        m_typeInfo(typeInfo)
    { readTypeInfo(typeInfo); }

    Interface (const IID &iid, const char *name):
        m_iid(iid),
        m_name(name)
    { }

public:
    // Get IID.
    const IID &iid () const
    { return m_iid; }

    // Get string representatin of IID.
    std::string iidString() const;

    // Get name.
    const std::string &name () const
    { return m_name; }

    // Get type info description.
    ITypeInfo *typeInfo () const
    { return m_typeInfo.get(); }

    // Insert method information.
    void addMethod(const Method &method);

    // Get methods.
    const Methods &methods () const
    { return m_methods; }

    // Find the named method.
    const Method *findMethod(const char *name) const;

    // Insert property information.
    void addProperty(const Property &property);

    // Get properties.
    const Properties &properties () const
    { return m_properties; }

    // Find the named property.
    const Property *findProperty(const char *name) const;

    // Return true if this interface derives from IDispatch.
    bool dispatchable () const
    { return m_dispatchable; }
};

// This is a cache of interface descriptions.

class TCOM_API InterfaceManager
{
    HashTable<IID, Interface *> m_hashTable;

    friend class Singleton<InterfaceManager>;
    static Singleton<InterfaceManager> ms_singleton;

    // Do not allow others to create or copy instances of this class.
    InterfaceManager();
    InterfaceManager(const InterfaceManager &);     // not implemented
    void operator=(const InterfaceManager &);       // not implemented

public:
    ~InterfaceManager();

    // Get singleton instance.
    static InterfaceManager &instance();

    // If the interface description already exists, return it,
    // otherwise create a new interface description.
    const Interface *newInterface(REFIID iid, TypeInfo &typeInfo);

    // If the interface description already exists, return it,
    // otherwise create a new interface description.
    Interface *newInterface(REFIID iid, const char *name);

    // Look for the interface description.
    const Interface *find(REFIID iid) const;
};

#endif 
