// $Id: TypeLib.cpp,v 1.16 2000/04/06 17:07:57 chuang Exp $
#include "TypeInfo.h"
#include "TypeLib.h"

Class::Class (TypeInfo &typeInfo,
              const Interface *pDefaultInterface,
              const Interface *pSourceInterface):
    m_name(typeInfo.name()),
    m_pDefaultInterface(pDefaultInterface),
    m_pSourceInterface(pSourceInterface)
{
    TypeAttr typeAttr(typeInfo);
    m_clsid = typeAttr->guid;
}


Enum::Enum (TypeInfo &typeInfo, TypeAttr &attr):
    m_name(typeInfo.name())
{
   HRESULT hr;

   for (int iEnum = 0; iEnum < attr->cVars; ++iEnum) {
       // Get enumerator description.
       VARDESC *pVarDesc;
       hr = typeInfo->GetVarDesc(iEnum, &pVarDesc);
       if (FAILED(hr)) {
           break;
       }

       // Get enumerator name.
       BSTR bstrName;
       UINT namesReturned;
       hr = typeInfo->GetNames(pVarDesc->memid, &bstrName, 1, &namesReturned);
       if (SUCCEEDED(hr)) {
           // Remember enumerator name and value.
           insert(value_type(std::string(_bstr_t(bstrName)),
               std::string(_bstr_t(_variant_t(pVarDesc->lpvarValue)))));
       }

       typeInfo->ReleaseVarDesc(pVarDesc);
   }
}


HRESULT
TypeLib::load (const char *name, TypeLib *&pTypeLib, bool registerTypeLib)
{
    _bstr_t nameStr(name);
    REGKIND regKind = registerTypeLib ? REGKIND_REGISTER : REGKIND_NONE;
    ITypeLib *pITypeLib;
    HRESULT hr = LoadTypeLibEx(nameStr, regKind, &pITypeLib);
    if (SUCCEEDED(hr)) {
        pTypeLib = new TypeLib(pITypeLib);
    }
    return hr;
}

std::string
TypeLib::name () const
{
    BSTR nameStr;
    HRESULT hResult = m_pITypeLib->GetDocumentation(
        MEMBERID_NIL, &nameStr, NULL, NULL, NULL);
    if (FAILED(hResult)) {
        return std::string();
    }
    _bstr_t wrapper(nameStr, false);
    return std::string(wrapper);
}

std::string
TypeLib::documentation () const
{
    BSTR docStr;
    HRESULT hResult = m_pITypeLib->GetDocumentation(
        MEMBERID_NIL, NULL, &docStr, NULL, NULL);
    if (FAILED(hResult)) {
        return std::string();
    }
    _bstr_t wrapper(docStr, false);
    return std::string(wrapper);
}

const Interface *
TypeLib::findInterface (const char *name) const
{
    Interfaces::const_iterator p = m_interfaces.begin();
    for (; p != m_interfaces.end(); ++p) {
        if ((*p)->name() == name) {
            return *p;
        }
    }
    return 0;
}

const Class *
TypeLib::findClass (const char *name) const
{
    Classes::const_iterator p = m_classes.begin();
    for (; p != m_classes.end(); ++p) {
        if (p->name() == name) {
            return &(*p);
        }
    }
    return 0;
}

const Enum *
TypeLib::findEnum (const char *name) const
{
    Enums::const_iterator p = m_enums.begin();
    for (; p != m_enums.end(); ++p) {
        if (p->name() == name) {
            return &(*p);
        }
    }
    return 0;
}

void
TypeLib::readTypeLib ()
{
    HRESULT hResult;

    unsigned count = m_pITypeLib->GetTypeInfoCount();
    for (unsigned index = 0; index < count; ++index) {
        ITypeInfo *pITypeInfo;
        hResult = m_pITypeLib->GetTypeInfo(index, &pITypeInfo);
        if (FAILED(hResult)) {
            continue;
        }
        TypeInfo typeInfo(pITypeInfo);
        TypeAttr typeAttr(typeInfo);

        switch (typeAttr->typekind) {
        case TKIND_DISPATCH:
        case TKIND_INTERFACE:
            // Read interface description.
            {
                const Interface *pInterface =
                    InterfaceManager::instance().newInterface(
                        typeAttr->guid, typeInfo);
                m_interfaces.push_back(pInterface);
            }
            break;

        case TKIND_COCLASS:
            // Read class description.
            {
                // Get the default interface for the class.
                ITypeInfo *pDefaultTypeInfo = typeInfo.getImplType(
                    IMPLTYPEFLAG_FDEFAULT);
                TypeInfo defaultTypeInfo(pDefaultTypeInfo);
                TypeAttr defaultTypeAttr(defaultTypeInfo);

                const Interface *pDefaultInterface =
                    InterfaceManager::instance().newInterface(
                        defaultTypeAttr->guid, defaultTypeInfo);

                // Get source interface for the class.
                const Interface *pSourceInterface;
                ITypeInfo *pSourceTypeInfo = typeInfo.getImplType(
                    IMPLTYPEFLAG_FDEFAULT | IMPLTYPEFLAG_FSOURCE);
                if (pSourceTypeInfo != 0) {
                    TypeInfo sourceTypeInfo(pSourceTypeInfo);
                    TypeAttr sourceTypeAttr(sourceTypeInfo);

                    pSourceInterface =
                        InterfaceManager::instance().newInterface(
                            sourceTypeAttr->guid, sourceTypeInfo);
                } else {
                    pSourceInterface = 0;
                }

                Class aClass(typeInfo, pDefaultInterface, pSourceInterface);
                m_classes.push_back(aClass);
            }
            break;

        case TKIND_ENUM:
            // Read the enumeration values.
            {
                Enum anEnum(typeInfo, typeAttr);
                m_enums.push_back(anEnum);
            }
            break;
        }
    }
}
