// $Id: objectCmd.cpp,v 1.18 2000/04/26 16:43:19 chuang Exp $
#include "objectCmd.h"
#include "ComObject.h"
#include "ComObjectFactory.h"
#include "ComModule.h"
#include "refCmd.h"
#include "importCmd.h"
#include "interfacePointer.h"

// This Tcl command registers a factory that creates COM objects.

static int
objectRegisterFactoryCmd (
    Tcl_Interp *interp,		/* Current interpreter. */
    int objc,			/* Number of arguments. */
    Tcl_Obj *CONST objv[])	/* The argument objects. */
{
    bool singletonOpt = false;

    int i = 2;
    for (; i < objc; ++i) {
        static char *options[] = {
	    "-singleton", NULL
        };
        enum OptionEnum {
            OPTION_SINGLETON
        };

        int index;
        if (Tcl_GetIndexFromObj(NULL, objv[i], options, "option", 0, &index)
         != TCL_OK) {
            break;
        }

        switch (index) {
        case OPTION_SINGLETON:
            singletonOpt = true;
            break;
        }
    }

    if (objc - i < 2 || objc - i > 3) {
	Tcl_WrongNumArgs(
            interp,
            2,
            objv,
            "?-singleton? class constructCommand ?destroyCommand?");
	return TCL_ERROR;
    }

    char *className = Tcl_GetStringFromObj(objv[i], 0);
    const ReferenceClass *pClass = ReferenceClass::find(interp, className);
    if (pClass == 0) {
        Tcl_AppendResult(interp, "unknown class ", className, NULL);
        return TCL_ERROR;
    }

    TclObject constructor(objv[i + 1]);

    TclObject destructor;
    if (objc - i == 3) {
        destructor = objv[i + 2];
    }

    ComObjectFactory *pFactory;
    if (singletonOpt) {
        pFactory = new SingletonObjectFactory(
            *(pClass->defaultInterface()),
            Tcl_GetCurrentThread(),
            interp,
            constructor,
            destructor);
    } else {
        pFactory = new ComObjectFactory(
            *(pClass->defaultInterface()),
            Tcl_GetCurrentThread(),
            interp,
            constructor,
            destructor);
    }
    ComModule::instance().registerFactory(pClass->clsid(), pFactory);
    return TCL_OK;
}


// This Tcl command connects a COM interface to a Tcl object command which
// implements the interface.

static int
objectCreateCmd (
    Tcl_Interp *interp,		/* Current interpreter. */
    int objc,			/* Number of arguments. */
    Tcl_Obj *CONST objv[])	/* The argument objects. */
{
    if (objc < 4 || objc > 5) {
	Tcl_WrongNumArgs(interp, 1, objv, "class servant ?destructor?");
	return TCL_ERROR;
    }

    char *className = Tcl_GetStringFromObj(objv[2], 0);
    const ReferenceClass *pClass = ReferenceClass::find(interp, className);
    if (pClass == 0) {
        Tcl_AppendResult(interp, "unknown class ", className, NULL);
        return TCL_ERROR;
    }

    TclObject servant(objv[3]);

    TclObject destructor;
    if (objc == 5) {
        destructor = objv[4];
    }

    try {
        IUnknown *pComObject = ComObject::newInstance(
            *(pClass->defaultInterface()),
            Tcl_GetCurrentThread(),
            interp,
            servant,
            destructor);

        Tcl_Obj *pObj = Tcl_NewObj();
        Tcl_InvalidateStringRep(pObj);
        pObj->typePtr = &interfacePointerObjType;
        pObj->internalRep.otherValuePtr =
            reinterpret_cast<ClientData>(pComObject);

        Tcl_SetObjResult(interp, pObj);
    }
    catch (_com_error &e) {
        return setComErrorResult(interp, e, __FILE__, __LINE__);
    }
    return TCL_OK;
}

// This Tcl command provides operations for creating COM objects.

int
objectCmd (
    ClientData clientData,	/* Not used. */
    Tcl_Interp *interp,		/* Current interpreter. */
    int objc,			/* Number of arguments. */
    Tcl_Obj *CONST objv[])	/* The argument objects. */
{
    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "subcommand ?arg ...?");
	return TCL_ERROR;
    }

    static char *options[] = {
	"create", "registerfactory", NULL
    };
    enum SubCommandEnum {
        CREATE, REGISTER_FACTORY
    };

    int index;
    if (Tcl_GetIndexFromObj(interp, objv[1], options, "subcommand", 0,
     &index) != TCL_OK) {
        return TCL_ERROR;
    }

    switch (index) {
    case CREATE:
        return objectCreateCmd(interp, objc, objv);
    case REGISTER_FACTORY:
        return objectRegisterFactoryCmd(interp, objc, objv);
    }
    return TCL_ERROR;
}
